# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function)

import os
from bothub_client.intent import IntentState
from bothub_client.dispatcher import DefaultDispatcher
from bothub_client.messages import Message


class BaseBot(object):
    '''A base Bot class'''

    def __init__(self, channel_client=None, storage_client=None, nlu_client_factory=None, event=None, dispatcher_class=None):
        '''Initialize an object

        :param channel_client: a ChannelClient object
        :type channel_client: bothub_client.clients.ChannelClient
        :param storage_client: a StorageClient object
        :type storage_client: bothub_client.clients.StorageClient
        :param nlu_client_factory: an NLU client factory object
        :type nlu_client_factory: bothub_client.clients.NluClientFactory
        :param event: an event which messenger platform sent
        :type event: dict'''
        self.channel_client = channel_client
        self.storage_client = storage_client
        self.nlu_client_factory = nlu_client_factory
        self.event = event
        bot_dir_path = os.path.realpath('.')
        yml_path = os.path.join(bot_dir_path, 'bothub.yml')
        if os.path.isfile(yml_path):
            self.intent_slots = IntentState.load_intent_slots_from_yml(yml_path)
        else:
            self.intent_slots = []

        self.state = IntentState(self, self.intent_slots)
        _dispatcher_class = dispatcher_class or DefaultDispatcher
        self.dispatcher = _dispatcher_class(self, self.state)

    def handle_message(self, event, context):
        '''Handle a message which messenger platform sent

        :param event: an event which messenger platform sent
        :type event: dict
        :param context: a context Bot runs
        :type context: dict'''

        self.dispatcher.dispatch(event, context)

    def send_message(self, message, chat_id=None, channel=None, extra=None, event=None):
        '''Send a message to an user or chatroom.

        :param message: a message to send. it can be a str text or Message class object
        :type message: str, bothub_client.messages.Message
        :param chat_id: receiver chat id
        :type chat_id: str
        :param channel: receiver channel name
        :type channel: str
        :param extra: deprecated
        :type extra: dict
        :return: None'''
        _event = event
        if _event is None and isinstance(message, Message):
            _event = message.event
        else:
            _event = self.event
        self.channel_client.send_message(chat_id, message, channel, event=_event, extra=extra)

    def send_photo(self, photo_url, chat_id=None, channel=None, event=None):
        '''Send a photo to an user or chatroom.

        :param photo_url: a photo to send.
        :type photo : str
        :param chat_id: receiver chat id
        :type chat_id: str
        :param channel: receiver channel name
        :type channel: str
        :return: None'''
        _event = self.event
        self.channel_client.send_photo(chat_id, photo_url, channel, event=_event)

    def set_project_data(self, data):
        '''Set project properties

        :param data: a dict to store
        :type data: dict
        :return: None'''
        self.storage_client.set_project_data(data)

    def get_project_data(self, key=None):
        '''Returns project properties

        :return: a properties dict
        :rtype: dict'''
        return self.storage_client.get_project_data(key)

    def set_user_data(self, data, user_id=None, channel=None, event=None):
        '''Set user properties

        :param data: a dict to store
        :type data: dict
        :param user_id: an user id to store data
        :type user_id: str, int
        :param channel: a name of messaging platform
        :type channel: str
        :return: None'''
        _user_id, _channel = self._get_channel_user(user_id, channel, event)
        self.storage_client.set_user_data(_channel, _user_id, data)

    def get_user_data(self, user_id=None, channel=None, event=None, key=None):
        '''Returns user properties

        :param user_id: an user id to store data
        :type user_id: str, int
        :param channel: a name of messaging platform
        :type channel: str
        :return: a properties dict
        :rtype: dict'''
        _user_id, _channel = self._get_channel_user(user_id, channel, event)
        return self.storage_client.get_user_data(_channel, user_id=_user_id, key=key)

    def nlu(self, vendor):
        '''Returns NLU client

        :param vendor: a NLU vendor name
        :type vendor: str
        :return: a NLU client
        :rtype: bothub_client.client.NluClient'''
        return self.nlu_client_factory.get(vendor)

    def _get_channel_user(self, user_id, channel, event):
        _event = event or self.event
        if user_id and channel:
            _user_id, _channel = user_id, channel
        else:
            _user_id, _channel = _event.get('sender', {}).get('id'), _event.get('channel')
        return _user_id, _channel

    def close(self):
        self.channel_client.close()



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)

import ast
import inspect
import sys
import traceback


def traceback_to_string(exc, tb=None):
    _tb = tb or exc.__traceback__
    s = traceback.extract_stack()[:-3] + traceback.extract_tb(_tb)
    l = traceback.format_list(s)
    return ''.join(l) + '\\n  {} {}'.format(exc.__class__, exc)


def get_decorators(cls):
    decorators = {}

    def visit_FunctionDef(node):
        decorators[node.name] = []
        for n in node.decorator_list:
            name = ''
            if isinstance(n, ast.Call):
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id

            args = [a.s for a in n.args] if hasattr(n, 'args') else []
            decorators[node.name].append((name, args))

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef
    _cls = cls if inspect.isclass(cls) else cls.__class__
    node_iter.visit(ast.parse(inspect.getsource(_cls)))
    return decorators



# -*- coding: utf-8 -*-

import requests_mock
from bothub_client.transports import HttpTransport
from bothub_client.transports import ZmqTransport


class DummySocket(object):
    def __init__(self):
        self.executed = []

    def connect(self, address):
        self.executed.append({'command': 'connect',
                              'args': {'address': address}})

    def send_json(self, data):
        self.executed.append({'command': 'send_json',
                              'args': {'data': data}})

    def send_multipart(self, data):
        self.executed.append({'command': 'send_multipart',
                              'args': {'data': data}})


class DummyContext(object):
    def __init__(self):
        self._socket = None

    def socket(self, sockettype):
        self.sockettype = sockettype
        if not self._socket:
            self._socket = DummySocket()
        return self._socket


def test_http_get_should_invoke_get():
    with requests_mock.mock() as mock:
        mock.get('localhost:8000/path', text='{"ran":true}')
        t = HttpTransport('localhost:8000')
        response = t.get('/path')
        assert response['ran'] is True
        assert mock.called is True


def test_http_post_should_invoke_post():
    with requests_mock.mock() as mock:
        mock.post('localhost:8000/path', text='{"ran":true}')
        t = HttpTransport('localhost:8000')
        t.post('/path', data='[]')
        assert mock.called is True


def test_http_put_should_invoke_put():
    with requests_mock.mock() as mock:
        mock.put('localhost:8000/path', text='{"ran":true}')
        t = HttpTransport('localhost:8000')
        t.put('/path', data='[]')
        assert mock.called is True


def test_zmq_send_json_should_invoke_send():
    context = DummyContext()
    t = ZmqTransport('localhost', context=context)
    t.send_json('mydata')
    assert context._socket.executed == [{'command': 'connect',
                                         'args': {'address': 'localhost'}},
                                        {'command': 'send_json',
                                         'args': {'data': 'mydata'}}]


def test_zmq_send_multipart_should_invoke_send():
    context = DummyContext()
    t = ZmqTransport('localhost', context=context)
    t.send_multipart('mydata')
    assert context._socket.executed == [{'command': 'connect',
                                         'args': {'address': 'localhost'}},
                                        {'command': 'send_multipart',
                                         'args': {'data': 'mydata'}}]



# -*- coding: utf-8 -*-

import sys
from bothub_client.utils import traceback_to_string
from bothub_client.utils import get_decorators
from bothub_client.decorators import command, intent


class TestDecorator(object):
    @command('start')
    def iamcommand(self):
        pass

    @command('reply')
    def iamcommand_too(self):
        pass

    @intent('credentials')
    def iamintent(self):
        pass


def test_get_decorators_should_returns_decorators():
    assert get_decorators(TestDecorator) == {'iamcommand': [('command', ['start'])],
                                             'iamcommand_too': [('command', ['reply'])],
                                             'iamintent': [('intent', ['credentials'])]}


def test_get_decorators_with_obj_should_returns_decorators():
    assert get_decorators(TestDecorator()) == {'iamcommand': [('command', ['start'])],
                                               'iamcommand_too': [('command', ['reply'])],
                                               'iamintent': [('intent', ['credentials'])]}


def test_traceback_to_string_should_return_string():
    try:
        raise KeyError()
    except KeyError as e:
        tb = None
        if sys.version_info < (3, 0):
            _, _, tb = sys.exc_info()
        s = traceback_to_string(e, tb=tb)
        assert 'raise KeyError()' in s



# -*- coding: utf-8 -*-

import json
from collections import namedtuple
from bothub_client.dispatcher import DefaultDispatcher
from bothub_client.intent import Intent
from bothub_client.intent import Slot
from bothub_client.intent import IntentState
from bothub_client.decorators import command
from bothub_client.decorators import intent
from bothub_client.decorators import channel

Executed = namedtuple('Executed', ['command', 'args'])


class MockOldStyleBot(object):
    def __init__(self):
        self.data = {}
        self.executed = []
        self.sent = []

    def get_user_data(self):
        return self.data

    def set_user_data(self, data):
        data_json_str = json.dumps(data)
        json_data = json.loads(data_json_str)
        self.data.update(**json_data)

    def on_default(self, *args):
        self.executed.append(Executed('on_default', args))

    def send_message(self, message):
        self.sent.append(message)

    def set_credentials(self, event, context, answers):
        self.executed.append(Executed('set_credentials', (answers['app_id'], answers['app_secret'])))

    def on_hello(self, event, context, *args):
        self.executed.append(Executed('hello', (event, context) + args))

    def on_no_args(self, event, context):
        self.executed.append(Executed('no_args', (event, context)))


class MockNewStyleBot(object):
    def __init__(self):
        self.data = {}
        self.executed = []
        self.sent = []

    def get_user_data(self):
        return self.data

    def set_user_data(self, data):
        data_json_str = json.dumps(data)
        json_data = json.loads(data_json_str)
        self.data.update(**json_data)

    @channel()
    def default(self, *args):
        self.executed.append(Executed('default', args))

    def send_message(self, message):
        self.sent.append(message)

    @intent('credentials')
    def set_credentials(self, event, context, answers):
        self.executed.append(Executed('set_credentials', (event, context, answers)))

    @command('hello')
    def hello(self, event, context, args):
        self.executed.append(Executed('hello', (event, context)))


def fixture_intent_slots():
    return [
        Intent('credentials', None, [
            Slot('app_id', 'Please tell me your app ID', None, 'string'),
            Slot('app_secret', 'Please tell me your app secret', None, 'string'),
        ]),
        Intent('address', None, [
            Slot('country', 'Please tell me your country', None, 'string'),
            Slot('city', 'Please tell me your city', None, 'string'),
            Slot('road', 'Please tell me your road address', None, 'string'),
        ])
    ]


def test_simple_message_dispatch_should_call_default_channel_handler():
    bot = MockNewStyleBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)

    event = {'content': 'hello', 'channel': 'fakechannel'}
    dispatcher.dispatch(event, None)

    executed = bot.executed.pop(0) # type: Executed
    assert executed.command == 'default'
    assert executed.args == (event, None)


def test_simple_message_dispatch_should_call_execute_command():
    bot = MockNewStyleBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)

    event = {'content': '/hello', 'channel': 'fakechannel'}
    dispatcher.dispatch(event, None)

    executed = bot.executed.pop(0) # type: Executed
    assert executed.command == 'hello'
    assert executed.args == (event, None)


def test_simple_message_dispatch_should_call_intent_complete_handler():
    bot = MockNewStyleBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)

    event1 = {'content': '/intent credentials', 'channel': 'fakechannel'}
    dispatcher.dispatch(event1, None)

    event2 = {'content': 'myid', 'channel': 'fakechannel'}
    dispatcher.dispatch(event2, None)
    event3 = {'content': 'mysecret', 'channel': 'fakechannel'}
    dispatcher.dispatch(event3, None)

    executed = bot.executed.pop(0) # type: Executed
    assert executed.command == 'set_credentials'
    assert executed.args == (event3, None, {'app_id': 'myid', 'app_secret': 'mysecret'})


def test_simple_message_old_style_dispatch_should_call_default_channel_handler():
    bot = MockOldStyleBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)

    event = {'content': 'hello', 'channel': 'fakechannel'}
    dispatcher.dispatch(event, None)

    executed = bot.executed.pop(0) # type: Executed
    assert executed.command == 'on_default'
    assert executed.args == (event, None)


def test_simple_message_old_style_dispatch_should_call_execute_command():
    bot = MockOldStyleBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)

    event = {'content': '/hello', 'channel': 'fakechannel'}
    dispatcher.dispatch(event, None)

    executed = bot.executed.pop(0) # type: Executed
    assert executed.command == 'hello'
    assert executed.args == (event, None)


def test_simple_message_old_style_dispatch_should_call_execute_command_without_args():
    bot = MockOldStyleBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)

    event = {'content': '/no_args', 'channel': 'fakechannel'}
    dispatcher.dispatch(event, None)

    executed = bot.executed.pop(0) # type: Executed
    assert executed.command == 'no_args'
    assert executed.args == (event, None)


def test_simple_message_old_style_dispatch_should_call_execute_command_with_args():
    bot = MockOldStyleBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)

    event = {'content': '/hello arg1', 'channel': 'fakechannel'}
    dispatcher.dispatch(event, None)

    executed = bot.executed.pop(0) # type: Executed
    assert executed.command == 'hello'
    assert executed.args == (event, None, 'arg1')



# -*- coding: utf-8 -*-

import json
import logging
from collections import namedtuple

import pytest
from bothub_client.intent import Intent
from bothub_client.intent import Slot
from bothub_client.intent import IntentState
from bothub_client.intent import NoSlotRemainsException
from bothub_client.dispatcher import DefaultDispatcher
from bothub_client.decorators import channel
from bothub_client.decorators import intent


Executed = namedtuple('Executed', ['command', 'args'])

class MockBot(object):
    def __init__(self):
        self.data = {}
        self.executed = []
        self.sent = []

    def get_user_data(self):
        return self.data

    def set_user_data(self, data):
        data_json_str = json.dumps(data)
        json_data = json.loads(data_json_str)
        self.data.update(**json_data)

    @channel()
    def on_default(self, *args):
        self.executed.append(Executed('on_default', args))

    def send_message(self, message):
        self.sent.append(message)

    @intent('credentials')
    def set_credentials(self, event, context, answers):
        self.executed.append(Executed('set_credentials', (answers['app_id'], answers['app_secret'])))


def fixture_intent_slots():
    return [
        Intent('credentials', None, [
            Slot('app_id', 'Please tell me your app ID', None, 'string'),
            Slot('app_secret', 'Please tell me your app secret', None, 'string'),
        ]),
        Intent('address', None, [
            Slot('country', 'Please tell me your country', None, 'string'),
            Slot('city', 'Please tell me your city', None, 'string'),
            Slot('road', 'Please tell me your road address', None, 'string'),
        ])
    ]


def test_init_intent_should_set_init_entries():
    bot = MockBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    state.open('credentials')
    assert bot.data['_intent_id'] == 'credentials'
    assert bot.data['_remaining_slots'] == [
        {'id': 'app_id', 'question': 'Please tell me your app ID', 'datatype': 'string', 'options': None},
        {'id': 'app_secret', 'question': 'Please tell me your app secret', 'datatype': 'string', 'options': None},
    ]


def test_get_result_should_return_result_with_next_message():
    bot = MockBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    state.open('credentials')
    assert bot.data['_remaining_slots'] == [
        {'id': 'app_id', 'question': 'Please tell me your app ID', 'datatype': 'string', 'options': None},
        {'id': 'app_secret', 'question': 'Please tell me your app secret', 'datatype': 'string', 'options': None},
    ]

    result = state.next()
    assert result.completed is False
    assert result.next_message == 'Please tell me your app ID'
    assert bot.data['_remaining_slots'] == [
        {'id': 'app_secret', 'question': 'Please tell me your app secret', 'datatype': 'string', 'options': None},
    ]

    result = state.next({'content': '<my app ID>'})
    assert result.completed is False
    assert result.next_message == 'Please tell me your app secret'
    assert bot.data['_remaining_slots'] == []
    assert result.answers == {
        'app_id': '<my app ID>'
    }

    result = state.next({'content': '<my app secret>'})
    assert result.completed is True
    assert result.next_message is None
    assert result.answers == {
        'app_id': '<my app ID>',
        'app_secret': '<my app secret>'
    }


def test_get_result_should_raise_exception_when_exceeded_slots():
    bot = MockBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    state.open('credentials')
    state.next() # Returns IntentResult: q=Please tell me your app ID
    state.next() # Returns IntentResult: q=Please tell me your app secret
    result = state.next() # Returns IntentResult: complete=True
    assert result.completed is True

    with pytest.raises(NoSlotRemainsException):
        state.next()


def test_dispatch_should_execute_default():
    bot = MockBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)
    dispatcher.dispatch({'content': 'hello', 'channel': 'fakechannel'}, None)
    assert len(bot.executed) == 1
    executed = bot.executed.pop(0)
    assert executed == Executed(
        'on_default',
        ({'content': 'hello', 'channel': 'fakechannel'}, None)
    )


def test_dispatch_should_execute_credentials():
    logging.basicConfig(level=logging.DEBUG)
    bot = MockBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)
    dispatcher.dispatch({'content': '/intent credentials'}, None)
    dispatcher.dispatch({'content': 'my token'}, None)
    dispatcher.dispatch({'content': 'my secret token'}, None)
    assert len(bot.executed) == 1
    executed = bot.executed.pop(0)
    assert executed == Executed('set_credentials', ('my token', 'my secret token'))


def test_dispatch_should_trigger_intent_and_default():
    logging.basicConfig(level=logging.DEBUG)
    bot = MockBot()
    intent_slots = fixture_intent_slots()
    state = IntentState(bot, intent_slots)
    dispatcher = DefaultDispatcher(bot, state)
    dispatcher.dispatch({'content': '/intent credentials', 'channel': 'fakechannel'}, None)
    dispatcher.dispatch({'content': 'my token', 'channel': 'fakechannel'}, None)
    dispatcher.dispatch({'content': 'my secret token', 'channel': 'fakechannel'}, None)
    dispatcher.dispatch({'content': 'hello', 'channel': 'fakechannel'}, None)
    assert len(bot.executed) == 2
    executed = bot.executed.pop(0)
    assert executed == Executed('set_credentials', ('my token', 'my secret token'))
    executed = bot.executed.pop(0)
    assert executed == Executed('on_default', ({'content': 'hello', 'channel': 'fakechannel'},
                                               None))


def test_intent_state_load_intent_slots_should_return_intent_slots():
    intent_slots = IntentState.load_intent_slots_from_yml('tests/fixtures/test_bothub.yml')
    assert intent_slots == [Intent('age', None, [Slot('age', 'How old are you?', [], 'string'),
                                                 Slot('name', 'What is your name?', [], 'string')])]



# -*- coding: utf-8 -*-

MARKUP_MARKDOWN = 'markdown'
MARKUP_HTML = 'html'


class Markdown(object):
    def __init__(self, text):
        self.text = text


class HTML(object):
    def __init__(self, text):
        self.text = text


class Message(object):
    '''A message class which renders rich messages like button, quick replies

    ex) Message(event).set_text('Hi there!')\\
                      .add_quick_reply('Say hello')'''
    def __init__(self, event):
        '''Initialize message class

        :param event: an event object
        :type event: dict'''
        self.model = []
        self.event = event

    def set_text(self, text):
        '''Set text message

        :param text: a text to send
        :type text: str
        :param markup: a markup
        :type markup: str'''

        if isinstance(text, Markdown):
            _text = text.text
            _markup = MARKUP_MARKDOWN
        elif isinstance(text, HTML):
            _text = text.text
            _markup = MARKUP_HTML
        else:
            _text = text
            _markup = None

        self.model.append({
            'command': 'set_text',
            'args': {
                'text': _text,
                'markup': _markup,
            }
        })
        return self

    def add_image(self, url, width, height):
        '''Add image

        :param url: image url
        :type url: str
        :param width: image's width
        :type width: str
        :param height: image's height
        :type height: str'''
        self.model.append({
            'command': 'add_image',
            'args': {
                'url': url,
                'width': width,
                'height': height
            }
        })
        return self

    def add_url_button(self, text, url):
        '''Add an URL button

        :param text: button text
        :type text: str
        :param url: URL to link
        :type url: str'''
        self.model.append({
            'command': 'add_url_button',
            'args': {
                'text': text,
                'url': url
            }
        })
        return self

    def add_postback_button(self, text, payload):
        '''Add a postback button

        :param text: button text
        :type text: str
        :param payload: a payload to returns when button is clicked
        :type payload: str'''
        self.model.append({
            'command': 'add_postback_button',
            'args': {
                'text': text,
                'payload': payload
            }
        })
        return self

    def add_quick_reply(self, text, payload=None, image_url=None):
        '''Add a quick reply button

        :param text: button text
        :type text: str
        :param payload: a payload to returns when button is clicked
        :type payload: str
        :param image_url: an icon image URL. Only works with Facebook Messenger
        :type image_url: str'''
        self.model.append({
            'command': 'add_quick_reply',
            'args': {
                'text': text,
                'image_url': image_url,
                'payload': payload,
            }
        })
        return self

    def add_location_request(self, text):
        '''Add a location request

        :param text: button text. Will be ignored with Facebook Messenger
        :type text: str'''
        self.model.append({
            'command': 'add_location_request',
            'args': {
                'text': text
            }
        })
        return self

    def add_keyboard_button(self, text):
        '''Add a keyboard button

        :param text: Keyboard button text. Not supported for Facebook Messenger for now
        :type text: str'''
        self.model.append({
            'command': 'add_keyboard_button',
            'args': {
                'text': text
            }
        })
        return self

    def add_template(self, payload):
        '''Add a template payload

        :param payload: template payload. Only Facebook Messenger supported.
        :type payload: dict'''
        self.model.append({
            'command': 'add_template',
            'args': {
                'payload': payload
            }
        })
        return self

    def __repr__(self):
        result = []
        for model_entry in self.model:
            result.append('{} {}'.format(model_entry['command'], model_entry['args']))
        return '\n'.join(result)



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)

import json
import requests_mock

from six import u
from bothub_client.bot import BaseBot
from bothub_client.clients import handle_message
from bothub_client.clients import BaseChannelClient
from bothub_client.clients import ZmqChannelClient
from bothub_client.messages import Message


class Bot(BaseBot):
    def handle_message(self, event, context):
        content = event.get('content')
        self.send_message('hello, you said: {}'.format(content))


class DummyZmqTransport(object):
    def __init__(self):
        self.sent = []

    def send_multipart(self, data):
        self.sent.append(data)


def test_handle_message_should_add_channel_cmd_buff():
    event = {'content': 'hi!',
             'channel': 'mychannel',
             'sender': {'id': 'abcd1234',
                        'name': 'myname'}}

    context = {'channel': {'endpoint': 'http://localhost'}}

    with requests_mock.mock() as m:
        m.post('http://localhost/messages')
        response = handle_message(event, context, Bot)
        assert response['response'] is None
        assert m.called
        assert m.request_history[0].json() == {
            'channel': None,
            'receiver': None,
            'message': 'hello, you said: hi!',
            'event': event,
            'extra': None,
            'context': {'api_key': '', 'project_id': None}
        }


def test_get_channel_obj_should_returns_channel():
    context = {'channel': {'channels': [{'type': 'mychannel'}]}}
    client = BaseChannelClient(10, 'myapikey', 'myurl', context=context)
    channel = client._get_channel_obj('mychannel')
    assert channel == {'type': 'mychannel'}


def test_get_channel_obj_should_returns_none():
    context = {'channel': {'channels': [{'type': 'mychannel'}]}}
    client = BaseChannelClient(10, 'myapikey', 'myurl', context=context)
    channel = client._get_channel_obj('mychannel2')
    assert channel is None


def test_prepare_payload_should_returns_message_data():
    context = {'channel': {'channels': [{'type': 'mychannel'}]}}
    event = {}
    client = BaseChannelClient(10, 'myapikey', 'myurl', context=context)
    message = Message(event)
    data = client._prepare_payload('myid', message, channel='mychannel', event=event)
    assert data == {
        'channel': {'type': 'mychannel'},
        'receiver': 'myid',
        'context': {'project_id': 10, 'api_key': 'myapikey'},
        'event': {},
        'extra': None,
        'message': {
            'model': [],
            'event': {}
        }
    }


def test_zmq_channel_client_init_client_should_return_client():
    context = {'project_id': 1,
               'api_key': 'mykey',
               'channel_endpoint': 'tcp://localhost:1010'}
    transport = DummyZmqTransport()
    client = ZmqChannelClient.init_client(context, transport=transport)
    assert client.api_key == 'mykey'
    assert client.project_id == 1


def test_zmq_channel_client_send_message_should_invoke_transport():
    context = {'project_id': 1,
               'channel': {'endpoint': 'tcp://localhost:1010',
                           'type': 'mychannel',
                           'channels': [{'type': 'mychannel'}]},
               'api_key': 'mykey',
               'channel_endpoint': 'tcp://localhost:1010'}
    event = {'chat_id': '1124',
             'channel': 'mychannel'}
    transport = DummyZmqTransport()
    client = ZmqChannelClient.init_client(context, transport=transport)
    client.send_message('1123', {'message': 'Hello'}, event=event)
    message = transport.sent.pop(0)
    message_str = message[0].decode('utf-8')
    message_json = json.loads(message_str)
    assert message_json == {'receiver': '1123',
                            'message': {'message': 'Hello'},
                            'event': {'chat_id': '1124',
                                      'channel': 'mychannel'},
                            'context': {'api_key': 'mykey',
                                        'project_id': 1},
                            'channel': {'type': 'mychannel'},
                            'extra': None}



# -*- coding: utf-8 -*-

import json
import pytest
from io import BytesIO
from bothub_client.clients import NluClient
from bothub_client.clients import NluAction
from bothub_client.clients import ApiAiNluClient


class MockApiAiRequest(object):
    def __init__(self):
        self.query = None
        self.session_id = None

    def getresponse(self):
        response = {'request': {'query': self.query,
                                'session_id': self.session_id},
                    'result': {'action': 'myaction',
                               'parameters': [],
                               'actionIncomplete': True,
                               'fulfillment': {'speech': 'next question'}}}
        return BytesIO(json.dumps(response).encode('utf8'))


class MockApiai(object):
    def text_request(self):
        return MockApiAiRequest()


def fixture_apiai_response():
    response_dict = {'result': {'action': 'myaction',
                                'parameters': [],
                                'actionIncomplete': True,
                                'fulfillment': {'speech': 'Next question?'}}}
    return response_dict


def test_nlu_client_ask_raises_exception():
    with pytest.raises(NotImplementedError):
        NluClient().ask()


def test_apiai_nlu_client_parse_should_return_response():
    response_dict = fixture_apiai_response()
    response = BytesIO(json.dumps(response_dict).encode('utf8'))
    result = ApiAiNluClient.parse_response(response)
    assert isinstance(result.action, NluAction)
    assert result.next_message == 'Next question?'


def test_apiai_nlu_client_ask_with_event_should_return_response():
    client = ApiAiNluClient('myapikey')
    client.apiai = MockApiai()
    response = client._ask_with_event({'content': 'hello',
                                       'channel': 'mychannel',
                                       'sender': {'id': 'myid'}})
    _assert_apiai_response(response)


def test_apiai_nlu_client_ask_with_message_should_return_response():
    client = ApiAiNluClient('myapikey')
    client.apiai = MockApiai()
    response = client._ask_with_message('hello', 'mychannel-myid')
    _assert_apiai_response(response)


def test_apiai_nlu_client_ask_event_should_return_response():
    client = ApiAiNluClient('myapikey')
    client.apiai = MockApiai()
    response = client.ask(event={'content': 'hello',
                                 'channel': 'mychannel',
                                 'sender': {'id': 'myid'}})
    _assert_apiai_response(response)


def test_apiai_nlu_client_ask_message_should_return_response():
    client = ApiAiNluClient('myapikey')
    client.apiai = MockApiai()
    response = client.ask(message='hello', session_id='mychannel-myid')
    _assert_apiai_response(response)


def _assert_apiai_response(response):
    assert response.raw_response['result'] == {'action': 'myaction',
                                               'parameters': [],
                                               'actionIncomplete': True,
                                               'fulfillment': {'speech': 'next question'}}
    assert response.raw_response['request']['query'] == 'hello'
    assert response.raw_response['request']['session_id'] == 'mychannel-myid'



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function, unicode_literals)

import requests_mock

from bothub_client.clients import StorageClient
from .testutils import MockTransport


def fixture_client(transport=None):
    context = {
        'project_id': 1,
        'api_key': 'testkey',
        'storage': {
            'endpoint': 'http://192.168.0.1:9000'
        }
    }

    client = StorageClient.init_client(context, transport=transport)
    return client


def record_project_data(transport):
    transport.record({'score': 11})


def test_init_client_should_returns_client_obj():
    client = fixture_client()
    assert client.project_id == 1
    assert client.api_key == 'testkey'
    assert client.base_url == 'http://192.168.0.1:9000'


def test_set_project_data():
    transport = MockTransport()

    client = fixture_client(transport)
    client.set_project_data({'score': 11})

    assert transport.executed.pop(0) == ('put',
                                         {'url': '/projects/1',
                                          'data': {'data': {'score': 11}}})


def test_get_project_data():
    transport = MockTransport()
    transport.record({})

    client = fixture_client(transport)
    client.get_project_data()

    assert transport.executed.pop(0) == ('get', {'url': '/projects/1'})


def test_set_user_data():
    transport = MockTransport()

    client = fixture_client(transport)
    client.set_user_data('mychannel', 'yourid', {'score': 11})

    assert transport.executed.pop(0) == ('put',
                                         {'url': '/projects/1/channels/mychannel/users/yourid',
                                          'data': {'data': {'score': 11}}})


def test_get_user_data():
    transport = MockTransport()
    transport.record({})

    client = fixture_client(transport)
    client.get_user_data('mychannel', 'yourid')

    assert transport.executed.pop(0) == ('get',
                                         {'url': '/projects/1/channels/mychannel/users/yourid'})



# -*- coding: utf-8 -*-

import logging
from bothub_client.utils import get_decorators
from bothub_client.messages import Message

logger = logging.getLogger('bothub.dispatcher')

class DefaultDispatcher(object):
    default_handler_name = 'on_default'
    command_handler_pattern = 'on_{command}'

    def __init__(self, bot, state):
        '''Initializer

        :param bot: a Bot object
        :type bot: BaseBot
        :param state: an IntentState object
        :type state: IntentState
        '''
        self.bot = bot
        self.state = state
        self.command_handlers = {}
        self.intent_handlers = {}
        self.channel_handlers = {}
        self._read_handlers()

    def _read_handlers(self):
        dec_type_to_handler_dict = {'command': self.command_handlers,
                                    'intent': self.intent_handlers,
                                    'channel': self.channel_handlers}
        method_to_decorators = get_decorators(self.bot)
        logger.debug('dispatch: method_to_decorators - %s', method_to_decorators)
        for method_name, decorators in method_to_decorators.items():
            for dec_type, args in decorators:
                handler_dict = dec_type_to_handler_dict.get(dec_type)
                if handler_dict is None:
                    continue
                handler_name = args[0] if args else 'default'
                handler_dict[handler_name] = method_name

    def dispatch(self, event, context):
        '''Dispatch incoming message event.

        :param event: an event object
        :type event: dict
        :param context: a context object
        :type content: dict
        '''
        logger.debug('dispatch: started')

        content = event.get('payload') or event.get('content')

        if self._is_intent_command(content):
            self.open_intent(event, content)
            return

        if self._is_command(content):
            self.execute_command(event, context, content)
            return

        if self.state.is_opened():
            self.proceed_intent(event, context)
            return

        current_channel = event.get('channel')
        if current_channel is None:
            return

        channel_handler = self.channel_handlers.get(current_channel)
        if channel_handler is None:
            channel_handler = self.channel_handlers.get('default', None)
        if channel_handler:
            handler_func = getattr(self.bot, channel_handler)
            handler_func(event, context)
        else:
            handler_func = getattr(self.bot, self.default_handler_name, None)
            if handler_func:
                handler_func(event, context)

    def open_intent(self, event, content):
        intent_id = self._get_intent_id(content)
        logger.debug('dispatch: intent %s started', intent_id)
        self.state.open(intent_id)
        result = self.state.next()
        message = Message(event)
        message.set_text(result.next_message)
        if result.options:
            for option in result.options:
                message.add_postback_button(option, option)
        self.bot.send_message(message)

    def execute_command(self, event, context, content):
        command, args = self._get_command_args(content)
        logger.debug('dispatch: start command %s', command)
        try:
            new_style_handler_name = self.command_handlers.get(command)
            old_style_handler_name = self.command_handler_pattern.format(command=command)

            if new_style_handler_name:
                handler_func = getattr(self.bot, new_style_handler_name)
                handler_func(event, context, args)
            elif old_style_handler_name:
                handler_func = getattr(self.bot, old_style_handler_name)
                handler_func(event, context, *args)

        except (KeyError, AttributeError):
            if self._is_kakao_init_keyboard_command(content):
                message = Message(event)
                message.add_keyboard_button('__init_keyboard__')
                self.bot.send_message(message)
            else:
                self.bot.send_message('No such command: {}'.format(command))

    def proceed_intent(self, event, context):
        logger.debug('dispatch: continue to process intent')
        result = self.state.next(event)
        if result.completed:
            logger.debug('dispatch: intent completed')

            new_sytle_handler_name = self.intent_handlers.get(result.intent_id)
            old_style_handler_name = result.complete_handler_name or 'set_{}'.format(result.intent_id)

            if new_sytle_handler_name:
                handler_func = getattr(self.bot, new_sytle_handler_name)
                handler_func(event, context, result.answers)
            elif old_style_handler_name:
                handler_func = getattr(self.bot, old_style_handler_name)
                handler_func(event, context, **result.answers)
        else:
            message = Message(event)
            message.set_text(result.next_message)
            if result.options:
                for option in result.options:
                    message.add_postback_button(option, option)
            self.bot.send_message(message)

    def _is_command(self, content):
        return content is not None and content.startswith('/')

    def _is_intent_command(self, content):
        return content is not None and content.startswith('/intent ')

    def _is_kakao_init_keyboard_command(self, content):
        return content is not None and content == '/kakao_init_keyboard'

    def _get_intent_id(self, content):
        _, intent_id = content.split()
        return intent_id

    def _get_command_args(self, content):
        tokens = content.split()
        command = tokens[0][1:]
        args = tokens[1:]
        return command, args



# -*- coding: utf-8 -*-

import logging
import yaml
from collections import namedtuple

logger = logging.getLogger('bothub.intent')

Intent = namedtuple('Intent', ['id', 'on_complete', 'slots'])
IntentResult = namedtuple('IntentResult', ['intent_id', 'completed', 'answers', 'next_message', 'complete_handler_name', 'options'])
Slot = namedtuple('Slot', ['id', 'question', 'options', 'datatype'])


class NoSlotRemainsException(Exception):
    pass


class IntentState(object):
    '''I manage intent state.

    Store and manage intent state in Bot user data storage.
    '''
    intent_id_field = '_intent_id'
    intent_answers_field = '_intent_answers'
    remaining_slots_field = '_remaining_slots'
    slot_id_field = '_slot_id'
    slot_datatype_field = '_slot_datatype'

    def __init__(self, bot, intent_slots):
        self.bot = bot
        self.intent_id_to_intent_definition = dict([(i.id, i) for i in intent_slots])

    @staticmethod
    def load_yml(path):
        with open(path, encoding='utf8') as fin:
            content = fin.read()
        return yaml.safe_load(content)

    @staticmethod
    def load_intent_slots_from_yml(path):
        config = IntentState.load_yml(path)
        intents = config.get('intents', {})
        intent_slots = []
        for intent_id in intents.keys():
            intent_yaml = intents[intent_id]
            slots_yaml = intent_yaml.get('slots', [])
            handler_name = intent_yaml.get('on_complete')
            slot_objs = []
            for slot in slots_yaml:
                _id = slot['id']
                question = slot['question']
                try:
                    options = slot['options']
                except KeyError:
                    options = []
                datatype = slot.get('datatype', 'string')
                slot_obj = Slot(_id, question, options, datatype)
                slot_objs.append(slot_obj)
            intent = Intent(intent_id, handler_name, slot_objs)
            intent_slots.append(intent)
        return intent_slots

    def open(self, intent_id):
        '''Open and start an intent.
        Should execute a `next()` method after open an intent.

        :param intent_id: an intent ID
        :type intent_id: str
        '''
        logger.debug('IntentState: init %s', intent_id)
        data = self.bot.get_user_data()
        data[self.intent_id_field] = intent_id
        data[self.intent_answers_field] = {}
        data[self.remaining_slots_field] = [
            d._asdict()
            for d in self.intent_id_to_intent_definition[intent_id].slots
        ]
        self.bot.set_user_data(data)

    def is_opened(self):
        '''Returns whether an intent is opened or not.

        :rtype: bool
        '''
        data = self.bot.get_user_data()
        return self.intent_id_field in data and data[self.intent_id_field] is not None

    def next(self, event=None):
        '''Proceed a current opened intent.
        If a user answer is given, store the answer and return an IntentResult object.

        :param event: an event dict
        :type event: dict
        :return: an intent result object
        :rtype: IntentResult
        '''
        logger.debug('IntentState: next with event %s', event)
        data = self.bot.get_user_data()
        if event:
            self._store_answer(event, data)
        result = self._make_result_obj(data)
        if result.completed:
            self._clear_state(data)
        self.bot.set_user_data(data)
        return result

    def close(self):
        '''Close an opened intent.

        Clear intent state variables in user data storage.
        '''
        data = self.bot.get_user_data()
        self._clear_state(data)
        self.bot.set_user_data(data)

    def _make_result_obj(self, data):
        next_message, options = self._next_slot_message(data)
        completed = next_message is None
        intent_id = data[self.intent_id_field]
        intent = self.intent_id_to_intent_definition[intent_id]
        result = IntentResult(intent_id,
                              completed,
                              dict(data[self.intent_answers_field].items()),
                              next_message,
                              intent.on_complete,
                              options)
        return result

    def _clear_state(self, data):
        logger.debug('IntentState: clear state')
        data[self.intent_id_field] = None
        data[self.slot_id_field] = None
        data[self.slot_datatype_field] = None
        data[self.remaining_slots_field] = None
        data[self.intent_answers_field] = None

    def _store_answer(self, event, data):
        slot_id = data[self.slot_id_field]
        data.setdefault(self.intent_answers_field, {})[slot_id] = event['content']

    def _has_remainig_slots(self, data):
        return self.remaining_slots_field in data and data[self.remaining_slots_field] is not None

    def _next_slot_message(self, data):
        try:
            if not self._has_remainig_slots(data):
                self.bot.set_user_data(data)
                raise NoSlotRemainsException()
            slot = data[self.remaining_slots_field].pop(0)
            data[self.slot_id_field] = slot['id']
            data[self.slot_datatype_field] = slot['datatype']
            return slot['question'], slot['options']
        except IndexError:
            return None, None



# -*- coding: utf-8 -*-

from bothub_client.messages import Message
from bothub_client.messages import Markdown
from bothub_client.messages import HTML


def test_set_text_should_append_entry():
    message = Message(None).set_text('hello')
    assert message.model == [{'command': 'set_text',
                              'args': {'text': 'hello', 'markup': None}}]


def test_set_text_with_markdown_should_append_entry():
    message = Message(None).set_text(Markdown('hello'))
    assert message.model == [{'command': 'set_text',
                              'args': {'text': 'hello', 'markup': 'markdown'}}]


def test_set_text_with_markdown_should_append_entry():
    message = Message(None).set_text(HTML('hello'))
    assert message.model == [{'command': 'set_text',
                              'args': {'text': 'hello', 'markup': 'html'}}]


def test_add_url_button_should_append_entry():
    message = Message(None).add_url_button('hello', 'myurl')
    assert message.model == [{'command': 'add_url_button',
                              'args': {'text': 'hello',
                                       'url': 'myurl'}}]


def test_add_postback_button_should_append_entry():
    message = Message(None).add_postback_button('hello', 'mypayload')
    assert message.model == [{'command': 'add_postback_button',
                              'args': {'text': 'hello',
                                       'payload': 'mypayload'}}]


def test_add_quick_reply_should_append_entry():
    message = Message(None).add_quick_reply('hello', 'mypayload', 'myurl')
    assert message.model == [{'command': 'add_quick_reply',
                              'args': {'text': 'hello',
                                       'payload': 'mypayload',
                                       'image_url': 'myurl'}}]


def test_add_location_request_should_append_entry():
    message = Message(None).add_location_request('hello')
    assert message.model == [{'command': 'add_location_request',
                              'args': {'text': 'hello'}}]


def test_add_keyboard_button_should_append_entry():
    message = Message(None).add_keyboard_button('hello')
    assert message.model == [{'command': 'add_keyboard_button',
                              'args': {'text': 'hello'}}]


def test_add_template_should_append_entry():
    message = Message(None).add_template('mypayload')
    assert message.model == [{'command': 'add_template',
                              'args': {'payload': 'mypayload'}}]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup

from codecs import open
from os import path

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.rst'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='bothub',
    version='0.1.30',
    description=u'Build serverless chatbot on BotHub.Studio',
    long_description=LONG_DESCRIPTION,
    author='Jeongsoo Park',
    author_email='jeongsoo.park@bothub.studio',
    url='https://github.com/bothub-studio/bothub-sdk-python',
    license="AGPL v3",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'requests',
        'zmq',
        'pyaml',
    ],
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'pytest',
        'requests-mock',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: Chat',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Utilities'
    ],
    keywords='bothub serverless chatbot framework',
)



# -*- coding: utf-8 -*-

from functools import wraps


def command(name):
    def dec(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return dec


def intent(name):
    def dec(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return dec


def channel(*names):
    def dec(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return dec



# -*- coding: utf-8 -*-

from bothub_client.bot import BaseBot


class DummyDispatcher(object):
    executed = []
    def __init__(self, bot, state):
        pass

    def dispatch(self, event, context):
        self.executed.append(('dispatch', {'event': event, 'context': context}))


class DummyStorageClient(object):
    def __init__(self):
        self.executed = []

    def set_project_data(self, data):
        self.executed.append(('set_project_data', {'data': data}))

    def get_project_data(self, key=None):
        self.executed.append(('get_project_data', {'key': key}))
        return {}

    def set_user_data(self, channel, user_id, data):
        self.executed.append(('set_user_data',
                              {'data': data, 'channel': channel, 'user_id': user_id}))

    def get_user_data(self, channel, user_id, key=None):
        self.executed.append(('get_user_data',
                              {'channel': channel, 'user_id': user_id, 'key': key}))
        return {}


class DummyNluClientFactory(object):
    def __init__(self):
        self.executed = []

    def get(self, vendor):
        self.executed.append(('get', {'vendor': vendor}))


def test_handle_message_should_run_dispatch():
    storage_client = DummyStorageClient()
    dispatcher_class = DummyDispatcher
    bot = BaseBot(storage_client=storage_client, event={}, dispatcher_class=dispatcher_class)
    bot.handle_message({}, {})
    assert dispatcher_class.executed.pop(0) == ('dispatch', {'event': {}, 'context': {}})


def test_set_project_data_should_invoke_client():
    storage_client = DummyStorageClient()
    bot = BaseBot(storage_client=storage_client, event={})
    bot.set_project_data({})
    assert storage_client.executed.pop(0) == ('set_project_data', {'data': {}})


def test_get_project_data_should_invoke_client():
    storage_client = DummyStorageClient()
    bot = BaseBot(storage_client=storage_client, event={})
    bot.get_project_data()
    assert storage_client.executed.pop(0) == ('get_project_data', {'key': None})


def test_set_user_data_should_invoke_client():
    storage_client = DummyStorageClient()
    bot = BaseBot(storage_client=storage_client, event={})
    bot.set_user_data({})
    assert storage_client.executed.pop(0) == ('set_user_data',
                                              {'data': {}, 'user_id': None, 'channel': None})


def test_get_user_data_should_invoke_client():
    storage_client = DummyStorageClient()
    bot = BaseBot(storage_client=storage_client, event={})
    bot.get_user_data()
    assert storage_client.executed.pop(0) == ('get_user_data',
                                              {'user_id': None, 'channel': None, 'key': None})


def test_nlu_should_return_client():
    nlu_factory = DummyNluClientFactory()
    bot = BaseBot(nlu_client_factory=nlu_factory, event={})
    bot.nlu('test')
    assert nlu_factory.executed.pop(0) == ('get', {'vendor': 'test'})
